#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <assert.h>

#include "include/htable.h"


#define BUCKET_EMPTY        1 << 0
#define BUCKET_DELETED      1 << 1

#if defined(DEBUG)
#define    ASSERT        assert
#else
#define    ASSERT(x)
#endif

#define    MAX_GROWTH_STEP          (1024U * 1024)

#define    APPROX_85_PERCENT(x)     (((x) * 870) >> 10)
#define    APPROX_40_PERCENT(x)     (((x) * 409) >> 10)


typedef struct {
    void *          key;
    void *          val;
    uint64_t        hash        : 32;
    uint64_t        psl         : 16;
} bucket_t;

typedef struct htable_t {
    unsigned        size;
    unsigned        nitems;
    unsigned        flags;
    uint64_t        divinfo;
    bucket_t *   buckets;
    uint64_t        hashkey;

    key_hash_fn_t   key_hash_fn;
    key_eq_fn_t     key_eq_fn;

    void *          key_data;
    ksize_t         ksize;
    void *          val_data;
    vsize_t         vsize;

    /*
     * Small optimisation for a single element case: allocate one
     * bucket together with the hashmap structure -- it will generally
     * fit within the same cache-line.
     */
    bucket_t    init_bucket;
} HTable;



/* * * GENERIC UTILITIES * * */

/*
 * Find first bit.
 */
static inline int fls(int x)
{ return x ? (sizeof(int) * CHAR_BIT) - __builtin_clz(x) : 0; }


/*
 * Fast 32bit division and remainder.
 *
 * Reference:
 *
 *    Torbjörn Granlund and Peter L. Montgomery, "Division by Invariant
 *    Integers Using Multiplication", ACM SIGPLAN Notices, Issue 6, Vol 29,
 *    http://gmplib.org/~tege/divcnst-pldi94.pdf, 61-72, June 1994.
 *
 * The following example computes q = a / b and r = a % b:
 *
 *    uint64_t divinfo = fast_div32_init(b);
 *    q = fast_div32(a, b, divinfo);
 *    r = fast_rem32(a, b, divinfo);
 */

static inline uint64_t
fast_div32_init(uint32_t div)
{
    uint64_t mt;
    uint8_t s1, s2;
    int l;

    l = fls(div - 1);
    mt = (uint64_t)(0x100000000ULL * ((1ULL << l) - div));
    s1 = (l > 1) ? 1U : (uint8_t)l;
    s2 = (l == 0) ? 0 : (uint8_t)(l - 1);
    return (uint64_t)(mt / div + 1) << 32 | (uint32_t)s1 << 8 | s2;
}

static inline uint32_t
fast_div32(uint32_t v, uint32_t div, uint64_t divinfo)
{
    const uint32_t m = divinfo >> 32;
    const unsigned s1 = (divinfo & 0x0000ff00) >> 8;
    const unsigned s2 = (divinfo & 0x000000ff);
    const uint32_t t = (uint32_t)(((uint64_t)v * m) >> 32);
    (void)div; // unused
    return (t + ((v - t) >> s1)) >> s2;
}


static inline uint32_t
fast_rem32(uint32_t v, uint32_t div, uint64_t divinfo)
{ return v - div * fast_div32(v, div, divinfo); }


//static int __attribute__((__unused__))
static int
validate_psl_p(HTable *ht, const bucket_t *bucket, unsigned i)
{
    unsigned base_i = fast_rem32(bucket->hash, ht->size, ht->divinfo);
    unsigned diff = (base_i > i) ? ht->size - base_i + i : i - base_i;
    return bucket->key == NULL || diff == bucket->psl;
}


/* * * PUBLIC API * * */

/*
 * Construct a new hash table.
 *
 * => If size is non-zero, then pre-allocate the given number of buckets;
 * => If size is zero, then a default minimum is used.
 */
int LSUP_htable_init(
        HTable *ht, htsize_t size, ksize_t ksize, vsize_t vsize,
        key_hash_fn_t key_hash_fn, key_eq_fn_t key_eq_fn, unsigned flags)
{
    ht->flags = flags;
    if (LSUP_htable_resize(ht, size) != 0) {
        free(ht);
        return -1;
    }
    ASSERT(ht->buckets);
    ASSERT(ht->size);

    return LSUP_OK;
}


HTable *LSUP_htable_new(
        htsize_t size, ksize_t ksize, vsize_t vsize,
        key_hash_fn_t key_hash_fn, key_eq_fn_t key_eq_fn, unsigned flags)
{
    HTable *ht;
    CRITICAL(ht = malloc(sizeof(HTable)));

    LSUP_htable_init(ht, size, ksize, vsize, key_hash_fn, key_eq_fn, flags);

    return ht;
}


/**
 * Resize a table.
 */
int LSUP_htable_resize(HTable *ht, htsize_t newsize)
{
    bucket_t *oldbuckets = ht->buckets;
    const size_t oldsize = ht->size;
    bucket_t *newbuckets;

    ASSERT(newsize > 0);
    ASSERT(newsize > ht->nitems);

    /*
     * Check for an overflow and allocate buckets.  Also, generate
     * a new hash key/seed every time we resize the hash table.
     */
    if (newsize == 1) {
        memset(&ht->init_bucket, 0, sizeof(bucket_t));
        newbuckets = &ht->init_bucket;
    } else if (newsize > UINT_MAX) {
        return -1;
    }

    CRITICAL(ht->buckets = calloc(1, newsize * sizeof(bucket_t)));
    CRITICAL(ht->key_data = realloc(ht->key_data, newsize * ht->ksize));
    CRITICAL(ht->val_data = realloc(ht->key_data, newsize * ht->vsize));

    ht->size = newsize;
    ht->nitems = 0;

    ht->divinfo = fast_div32_init(newsize);
    ht->hashkey ^= random() | (random() << 32);

    for (unsigned i = 0; i < oldsize; i++) {
        const bucket_t *bucket = &oldbuckets[i];

        /* Skip the empty buckets. */
        if (!bucket->key) {
            continue;
        }
        LSUP_htable_insert(ht, bucket->key, bucket->val);
        if ((ht->flags & HTABLE_NOCOPY) == 0) {
            free(bucket->key);
        }
    }
    if (oldbuckets && oldbuckets != &ht->init_bucket) {
        free(oldbuckets);
    }
    return 0;
}


htsize_t LSUP_htable_capacity(LSUP_HTable *ht)
{ return ht->size; }


htsize_t LSUP_htable_size(LSUP_HTable *ht)
{ return ht->nitems; }


/*
 * Insert without resizing (assuming resizing is already done).
 */
int LSUP_htable_insert(HTable *ht, const void *key, void *val)
{
    const uint32_t hash = ht->key_hash_fn(key, ht->ksize, ht->hashkey);
    bucket_t *bucket, entry;
    unsigned i;

    ASSERT(key != NULL);

    /*
     * Setup the bucket entry.
     */
    if (ht->flags & HTABLE_NOCOPY) {
        entry.key = (void *)(uintptr_t)key;
    } else {
        CRITICAL(entry.key = malloc(ht->ksize))
        memcpy(entry.key, key, ht->ksize);
    }
    entry.hash = hash;
    entry.val = val;
    entry.psl = 0;

    /*
     * From the paper: "when inserting, if a record probes a location
     * that is already occupied, the record that has traveled longer
     * in its probe sequence keeps the location, and the other one
     * continues on its probe sequence" (page 12).
     *
     * Basically: if the probe sequence length (PSL) of the element
     * being inserted is greater than PSL of the element in the bucket,
     * then swap them and continue.
     */
    i = fast_rem32(hash, ht->size, ht->divinfo);

    while(1) {
        bucket = &ht->buckets[i];
        if (bucket->key) {
            ASSERT(validate_psl_p(ht, bucket, i));

            // There is a key in the bucket.
            if (bucket->hash == hash && ht->key_eq_fn(
                        bucket->key, key, ht->ksize)) {
                // Duplicate key: do nothing.
                if ((ht->flags & HTABLE_NOCOPY) == 0) {
                    free(entry.key);
                }
                return LSUP_NOACTION;
            }

            /*
             * We found a "rich" bucket.  Capture its location.
             */
            if (entry.psl > bucket->psl) {
                bucket_t tmp;

                /*
                 * Place our key-value pair by swapping the "rich"
                 * bucket with our entry.  Copy the structures.
                 */
                tmp = entry;
                entry = *bucket;
                *bucket = tmp;
            }
            entry.psl++;

            /* Continue to the next bucket. */
            ASSERT(validate_psl_p(ht, bucket, i));
            i = fast_rem32(i + 1, ht->size, ht->divinfo);
        }
    }

    /*
     * Found a free bucket: insert the entry.
     */
    *bucket = entry; // copy
    ht->nitems++;

    ASSERT(validate_psl_p(ht, bucket, i));

    return LSUP_OK;
}


/*
 * rhashmap_put: insert a value given the key.
 *
 * => If the key is already present, return its associated value.
 * => Otherwise, on successful insert, return the given value.
 */
int LSUP_htable_put(HTable *ht, const void *key, void *val)
{
    const size_t threshold = APPROX_85_PERCENT(ht->size);

    /*
     * If the load factor is more than the threshold, then resize.
     */
    if (UNLIKELY(ht->nitems > threshold)) {
        /*
         * Grow the hash table by doubling its size, but with
         * a limit of MAX_GROWTH_STEP.
         */
        const size_t grow_limit = ht->size + MAX_GROWTH_STEP;
        const size_t newsize = min(ht->size << 1, grow_limit);
        LSUP_htable_resize(ht, newsize);
    }

    return LSUP_htable_insert(ht, key, val);
}


int LSUP_htable_get(const HTable *ht, const void *key, void **valp)
{
    const uint64_t hash = ht->key_hash_fn(key, ht->ksize, ht->hashkey);
    htsize_t n = 0, i = fast_rem32(hash, ht->size, ht->divinfo);

    if (key == NULL) return LSUP_VALUE_ERR;

    /*
     * Lookup is a linear probe.
     */
    while (1) {
        bucket_t *bucket = &ht->buckets[i];
        ASSERT(validate_psl_p(ht, bucket, i));

        if (bucket->hash == hash && ht->key_eq_fn(
                    bucket->key, key, ht->ksize))
            if (valp != NULL)
                *valp = ht->vsize == 0 ? NULL : bucket->val;

            return LSUP_OK;

        /*
         * Stop probing if we hit an empty bucket; also, if we hit a
         * bucket with PSL lower than the distance from the base location,
         * then it means that we found the "rich" bucket which should
         * have been captured, if the key was inserted -- see the central
         * point of the algorithm in the insertion function.
         */
        if (!bucket->key || n > bucket->psl) {
            valp = NULL;
            return LSUP_NORESULT;
        }
        n++;

        /* Continue to the next bucket. */
        i = fast_rem32(i + 1, ht->size, ht->divinfo);
    }
}


int LSUP_htable_del(HTable *ht, const void *key)
{
    const size_t threshold = APPROX_40_PERCENT(ht->size);
    const uint32_t hash = ht->key_hash_fn(key, ht->ksize, ht->hashkey);
    unsigned n = 0, i = fast_rem32(hash, ht->size, ht->divinfo);
    bucket_t *bucket;

    ASSERT(key != NULL);

    while(1) {
        /*
         * The same probing logic as in the lookup function.
         */
        bucket = &ht->buckets[i];
        if (!bucket->key || n > bucket->psl) {
            return LSUP_NOACTION;
        }
        ASSERT(validate_psl_p(ht, bucket, i));

        if (
                bucket->hash != hash ||
                ht->key_eq_fn(bucket->key, key, ht->ksize)) {
            /* Continue to the next bucket. */
            i = fast_rem32(i + 1, ht->size, ht->divinfo);
            n++;
        }
    }

    /*
     * Free the bucket.
     */
    if ((ht->flags & HTABLE_NOCOPY) == 0) {
        free(bucket->key);
    }
    ht->nitems--;

    /*
     * The probe sequence must be preserved in the deletion case.
     * Use the backwards-shifting method to maintain low variance.
     */
    while(1) {
        bucket_t *nbucket;

        bucket->key = NULL;

        i = fast_rem32(i + 1, ht->size, ht->divinfo);
        nbucket = &ht->buckets[i];
        ASSERT(validate_psl_p(ht, nbucket, i));

        /*
         * Stop if we reach an empty bucket or hit a key which
         * is in its base (original) location.
         */
        if (!nbucket->key || nbucket->psl == 0) {
            break;
        }

        nbucket->psl--;
        *bucket = *nbucket;
        bucket = nbucket;
    }

    /*
     * If the load factor is less than threshold, then shrink by
     * halving the size, but not less than 1.
     */
    if (ht->nitems < threshold) {
        size_t newsize = max(ht->size >> 1, 1);
        (void)LSUP_htable_resize(ht, newsize);
    }

    return LSUP_OK;
}


extern int LSUP_htable_iter(
        LSUP_HTable *ht, htsize_t *cur, void *key, void **valp)
{
    while (*cur < ht->size) {
        bucket_t *bucket = &ht->buckets[*cur];

        *cur++;

        if (!bucket->key) continue;

        memcpy(key, bucket->key, ht->ksize);
        if (valp != NULL && ht->vsize > 0) *valp = bucket->val;

        return LSUP_OK;
    }
    return LSUP_END;
}


void LSUP_htable_done(HTable *ht)
{
    if ((ht->flags & HTABLE_NOCOPY) == 0) {
        for (htsize_t i = 0; i < ht->size; i++) {
            const bucket_t *bucket = &ht->buckets[i];

            if (bucket->key) {
                free(bucket->key);
            }
        }
    }
    if (ht->buckets != &ht->init_bucket) {
        free(ht->buckets);
    }
}


void LSUP_htable_free(HTable *ht)
{
    if(LIKELY(ht != NULL)) {
        LSUP_htable_done(ht);
        free(ht);
    }
}
